/* This file is part of Legendes II */
/* Copyright (c) 1998 by
 *   Philippe Clavel (philippe.clavel@supelec.fr)
 *   David Madore (david.madore@ens.fr) */

/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
 * the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* File ViewArea.h */
/* This is the implementation file of the ViewArea widget. */

#include <stdio.h>

#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>

#include "ViewAreaP.h"

#define DEFAULT_ARRAY_WIDTH 32
#define DEFAULT_ARRAY_HEIGHT 32

#define DEFAULT_CELL_SIZE 16

#define offset(field) XtOffsetOf(ViewAreaRec,field)
static XtResource resources[] = {
  {XtNcolor0,XtCColor0,XtRPixel,sizeof(Pixel),
   offset(viewArea.color0),XtRString,"Black"},
  {XtNcolor1,XtCColor1,XtRPixel,sizeof(Pixel),
   offset(viewArea.color1),XtRString,"Grey51"},
  {XtNcolor2,XtCColor2,XtRPixel,sizeof(Pixel),
   offset(viewArea.color2),XtRString,"Red"},
  {XtNcallback,XtCCallback,XtRCallback,sizeof(XtPointer),
   offset(viewArea.callback),XtRCallback,NULL},
  {XtNarrayWidthInCells,XtCArrayWidthInCells,XtRDimension,sizeof(Dimension),
   offset(viewArea.array_width_in_cells),
   XtRImmediate,(XtPointer)DEFAULT_ARRAY_WIDTH},
  {XtNarrayHeightInCells,XtCArrayHeightInCells,XtRDimension,sizeof(Dimension),
   offset(viewArea.array_height_in_cells),
   XtRImmediate,(XtPointer)DEFAULT_ARRAY_HEIGHT},
  {XtNcellSizeInPixels,XtCCellSizeInPixels,XtRInt,sizeof(int),
   offset(viewArea.cell_size),
   XtRImmediate,(XtPointer)DEFAULT_CELL_SIZE},
  {XtNarrayData,XtCArrayData,XtRString,sizeof(int),
   offset(viewArea.cell),XtRImmediate,NULL}
};
#undef offset

static ViewAreaNotifyInfo info;

static void Notify(Widget,XEvent *,String *,Cardinal *);

static XtActionsRec actions[] = {
  {"notify",Notify}
};

static char defaultTranslations[] =
"<Btn1Down>: notify()\n\
<Btn3Down>: notify()";

/* The method prototypes */
static void Initialize(Widget,Widget,ArgList,Cardinal *);
static void Redisplay(Widget,XEvent *,Region);
static void Destroy(Widget);
static void Resize(Widget);
static Boolean SetValues(Widget,Widget,Widget,ArgList,Cardinal *);
static XtGeometryResult QueryGeometry(Widget,XtWidgetGeometry *,
				      XtWidgetGeometry *);

#define SuperClass ((CoreWidgetClass)&coreClassRec)

ViewAreaClassRec viewAreaClassRec = {
  { /* core */
    (WidgetClass)SuperClass,            /* superclass */
    "ViewArea",                         /* class_name */
    sizeof(ViewAreaRec),                /* widget_size */
    NULL,                               /* class_initialize */
    NULL,                               /* class_part_initialize */
    False,                              /* class_inited */
    Initialize,                         /* initialize */
    NULL,                               /* initialize_hook */
    XtInheritRealize,                   /* realize */
    actions,                            /* actions */
    XtNumber(actions),                  /* num_actions */
    resources,                          /* resources */
    XtNumber(resources),                /* num_resources */
    NULLQUARK,                          /* xrm_class */
    True,                               /* compress_motion */
    True,                               /* compress_exposure */
    True,                               /* compress_interleave */
    False,                              /* visible_interest */
    Destroy,                            /* destroy */
    Resize,                             /* resize */
    Redisplay,                          /* expose */
    SetValues,                          /* set_values */
    NULL,                               /* set_values_hook */
    XtInheritSetValuesAlmost,           /* set_values_almost */
    NULL,                               /* get_values_hook */
    NULL,                               /* accept_focus */
    XtVersion,                          /* version */
    NULL,                               /* callback_private */
    defaultTranslations,                /* tm_table */
    QueryGeometry,                      /* query_geometry */
    XtInheritDisplayAccelerator,        /* display_accelerator */
    NULL                                /* extension */
  },
  { /* viewArea */
    0                                   /* dummy */
  }
};

WidgetClass viewAreaWidgetClass = (WidgetClass)&viewAreaClassRec;

static void
Initialize(Widget request,Widget init,ArgList args,Cardinal *num_args)
{
  ViewAreaWidget w = (ViewAreaWidget)init; /* I love C! */
  if (w->viewArea.array_width_in_cells<1) {
    XtWarning("ViewArea: arrayWidth is too small (using 1).");
    w->viewArea.array_width_in_cells=1;
  }
  if (w->viewArea.array_height_in_cells<1) {
    XtWarning("ViewArea: arrayHeight is too small (using 1).");
    w->viewArea.array_height_in_cells=1;
  }
  if (w->viewArea.cell) {
    w->viewArea.cell_belongs_to_user=1;
  } else {
    w->viewArea.cell_belongs_to_user=0;
    w->viewArea.cell=XtCalloc(w->viewArea.array_height_in_cells
			       *w->viewArea.array_width_in_cells,
			       sizeof(char));
  }
  if (w->viewArea.cell_size<4) {
    XtWarning("ViewArea: cellSize is too small (using 4).");
    w->viewArea.cell_size=4;
  }
  w->viewArea.array_width_in_pixels=w->viewArea.array_width_in_cells
    *w->viewArea.cell_size;
  w->viewArea.array_height_in_pixels=w->viewArea.array_height_in_cells
    *w->viewArea.cell_size;
  w->core.width=w->viewArea.array_width_in_pixels;
  w->core.height=w->viewArea.array_height_in_pixels;
#if 0
  (*XtClass(w)->core_class.resize)(Widget(w)); /* ??? */
#endif
  {
    XGCValues values;
    XtGCMask mask = GCForeground | GCBackground;
    values.background = w->core.background_pixel;
    values.foreground = w->viewArea.color0;
    w->viewArea.gc_c0 = XtGetGC(init,mask,&values);
    values.foreground = w->viewArea.color1;
    w->viewArea.gc_c1 = XtGetGC(init,mask,&values);
    values.foreground = w->viewArea.color2;
    w->viewArea.gc_c2 = XtGetGC(init,mask,&values);
  }
}

char *
ViewAreaGetArrayData(Widget w0)
{
  ViewAreaWidget w = (ViewAreaWidget)w0;
  return w->viewArea.cell;
}

void
ViewAreaRedrawCell(Widget w0,int x,int y)
{
  ViewAreaWidget w = (ViewAreaWidget)w0;
  if (!XtIsRealized(w0)) return;
  if ((x<0)||(y<0)||(x>=w->viewArea.array_width_in_cells)
      ||(y>=w->viewArea.array_height_in_cells)) return;
  if (w->viewArea.cell[x*w->viewArea.array_height_in_cells+y])
    XFillRectangle(XtDisplay(w0),XtWindow(w0),
		   w->viewArea.gc_c1,
		   x*w->viewArea.cell_size,y*w->viewArea.cell_size,
		   w->viewArea.cell_size,w->viewArea.cell_size);
  else
    XFillRectangle(XtDisplay(w0),XtWindow(w0),
		   w->viewArea.gc_c0,
		   x*w->viewArea.cell_size,y*w->viewArea.cell_size,
		   w->viewArea.cell_size,w->viewArea.cell_size);
  if ((x==16)&&(y==16))
    XFillArc(XtDisplay(w0),XtWindow(w0),
	     w->viewArea.gc_c2,
	     x*w->viewArea.cell_size,y*w->viewArea.cell_size,
	     w->viewArea.cell_size,w->viewArea.cell_size,
	     0,23040);
}

void
ViewAreaRedrawCells(Widget w0,int x1,int y1,int x2,int y2)
{
  ViewAreaWidget w = (ViewAreaWidget)w0;
  int i,j;
  if (!XtIsRealized(w0)) return;
  if ((x2<x1)||(y2<y1)) return;
  if (x1<0) x1=0;
  if (y1<0) y1=0;
  if (x2>=w->viewArea.array_width_in_cells)
    x2=w->viewArea.array_width_in_cells-1;
  if (y2>=w->viewArea.array_height_in_cells)
    y2=w->viewArea.array_height_in_cells-1;
  for (i=x1;i<=x2;i++) for (j=y1;j<=y2;j++)
    ViewAreaRedrawCell(w0,i,j);
}

static void
Redisplay(Widget w0,XEvent *event,Region region)
{
  ViewAreaWidget w = (ViewAreaWidget)w0;
  ViewAreaRedrawCells(w0,0,0,w->viewArea.array_width_in_cells-1,
		       w->viewArea.array_height_in_cells-1);
}

static void
Destroy(Widget w0)
{
  ViewAreaWidget w = (ViewAreaWidget)w0;
  XtReleaseGC(w0,w->viewArea.gc_c0);
  XtReleaseGC(w0,w->viewArea.gc_c1);
  if (!w->viewArea.cell_belongs_to_user)
    XtFree(w->viewArea.cell);
}

static void
Resize(Widget w0)
{
  /* HOW DARE YOU RESIZE MY WIDGET YOU STUPID X TOOLKIT */
}

static Boolean
SetValues(Widget current,Widget request,Widget new,ArgList args,
	  Cardinal *num_args)
{
  ViewAreaWidget curw = (ViewAreaWidget)current;
  ViewAreaWidget neww = (ViewAreaWidget)new;
  if (neww->viewArea.array_width_in_cells!=
      curw->viewArea.array_width_in_cells) {
    XtWarning("ViewArea: arrayWidth is read-only.");
    neww->viewArea.array_width_in_cells
      =curw->viewArea.array_width_in_cells;
  }
  if (neww->viewArea.array_height_in_cells!=
      curw->viewArea.array_height_in_cells) {
    XtWarning("ViewArea: arrayHeight is read-only.");
    neww->viewArea.array_height_in_cells
      =curw->viewArea.array_height_in_cells;
  }
  if (neww->viewArea.cell_size<4) {
    XtWarning("ViewArea: cellSize is too small (using 4).");
    neww->viewArea.cell_size=4;
  }
  neww->viewArea.array_width_in_pixels=neww->viewArea.array_width_in_cells
    *neww->viewArea.cell_size;
  neww->viewArea.array_height_in_pixels=neww->viewArea.array_height_in_cells
    *neww->viewArea.cell_size;
  if ((neww->core.background_pixel!=curw->core.background_pixel)
      ||(neww->viewArea.color0!=curw->viewArea.color0)
      ||(neww->viewArea.color1!=curw->viewArea.color1)
      ||(neww->viewArea.color2!=curw->viewArea.color2)) {
    XGCValues values;
    XtGCMask mask = GCForeground | GCBackground;
    XtReleaseGC(new,curw->viewArea.gc_c0);
    XtReleaseGC(new,curw->viewArea.gc_c1);
    XtReleaseGC(new,curw->viewArea.gc_c2);
    values.background = neww->core.background_pixel;
    values.foreground = neww->viewArea.color0;
    neww->viewArea.gc_c0 = XtGetGC(new,mask,&values);
    values.foreground = neww->viewArea.color1;
    neww->viewArea.gc_c1 = XtGetGC(new,mask,&values);
    values.foreground = neww->viewArea.color2;
    neww->viewArea.gc_c2 = XtGetGC(new,mask,&values);
  }
  if (neww->viewArea.cell!=curw->viewArea.cell) {
    XtWarning("ViewArea: arrayData is read-only.");
    neww->viewArea.cell=curw->viewArea.cell;
  }
  return True;
}

static XtGeometryResult
QueryGeometry(Widget w0,XtWidgetGeometry *proposed,XtWidgetGeometry *answer)
{
  ViewAreaWidget w = (ViewAreaWidget)w0;
  answer->request_mode=CWWidth|CWHeight;
  answer->width=w->viewArea.array_width_in_pixels;
  answer->height=w->viewArea.array_height_in_pixels;
  if (((proposed->request_mode&(CWWidth|CWHeight))==(CWWidth|CWHeight))&&
      (proposed->width==answer->width)&&(proposed->height==answer->height))
    return XtGeometryYes;
  else if ((answer->width==w->core.width)&&(answer->height==w->core.height))
    return XtGeometryNo;
  else return XtGeometryAlmost;
}

static void
Notify(Widget w0,XEvent *event,String *params,Cardinal *num_params)
{
  ViewAreaWidget w = (ViewAreaWidget)w0;
  if ((event->type==ButtonPress)||(event->type==ButtonRelease)) {
    info.x=event->xbutton.x/w->viewArea.cell_size;
    info.y=event->xbutton.y/w->viewArea.cell_size;
  } else if ((event->type==KeyPress)||(event->type==KeyRelease)) {
    info.x=event->xkey.x/w->viewArea.cell_size;
    info.y=event->xkey.y/w->viewArea.cell_size;
  } else {
    info.x=0; /* Actually, they are undefined */
    info.y=0;
  }
  info.event=event;
  info.params=params;
  info.num_params=num_params;
  XtCallCallbacks(w0,XtNcallback,&info);
}

/* EOF */
