/* This file is part of Legendes II */
/* Copyright (c) 1998 by
 *   Philippe Clavel (philippe.clavel@supelec.fr)
 *   David Madore (david.madore@ens.fr) */

/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
 * the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* File input.c */
/* This file contains the input handling thread. */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <X11/Xlib.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Xaw/Box.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/Command.h>

#include "lgd0.h"
#include "lgdlib.h"
#include "fifo.h"

#include "ViewArea.h"

#include "input.h"
#include "output.h"
#include "legendes.h"

char input_inited;

pthread_t input_thread;

static char *fallback_resources[] = {
  "*input: True",
  "*background: Grey51",
  "*foreground: White",
  "*groundBox.background: DarkSlateGrey",
  "*groundBox.cuteLabel.foreground: Blue",
  "*groundBox.quitButton.label: Quit",
  "*groundBox.viewArea.translations: \\\n\
     <Key>Left: notify(move_west) \\n\\\n\
     <Key>Right: notify(move_east) \\n\\\n\
     <Key>Up: notify(move_north) \\n\\\n\
     <Key>Down: notify(move_south) \\n\\\n\
     <Btn1Down>: notify(write_square,1) \\n\\\n\
     <Btn3Down>: notify(write_square,0) \\n\\\n\
     <Key>F1: notify(write_chunk,0) \\n\\\n\
     <Key>F2: notify(write_chunk,1) \\n\\\n\
     <Key>F3: notify(write_chunk,2) \\n\\\n\
     <Key>F4: notify(write_chunk,3) \\n\\\n\
     <Key>F5: notify(write_chunk,4) \\n\\\n\
     <Key>F6: notify(write_chunk,5) \\n\\\n\
     <Key>F7: notify(write_chunk,6) \\n\\\n\
     <Key>F8: notify(write_chunk,7) \\n\\\n\
     <Key>F9: notify(write_chunk,8) \\n\\\n\
     <Key>F10: notify(write_chunk,9) \\n\\\n\
     <Key>F11: notify(write_chunk,10) \\n\\\n\
     <Key>F12: notify(write_chunk,10)",
  NULL
};

Widget toplevel,ground_box,cute_label,quit_button,view_area;

static XtAppContext app_context;

static msg_fifo_t msg;

static void
quit_callback(Widget what,XtPointer client_data,XtPointer call_data)
{
  exit(0);
}

static void
view_callback(Widget what,XtPointer client_data,XtPointer call_data)
{
  ViewAreaNotifyInfo *info = (ViewAreaNotifyInfo *)call_data;
  if (*info->num_params>=1) {
    if (strcmp(info->params[0],"move_west")==0) {
      while (!msg.treated) sched_yield();
      msg.type=msg_move_hero;
      msg.response=NULL;
      msg.param.move_hero.dx=-1;
      msg.param.move_hero.dy=0;
      append_control_msg(&msg);
    } else if (strcmp(info->params[0],"move_east")==0) {
      while (!msg.treated) sched_yield();
      msg.type=msg_move_hero;
      msg.response=NULL;
      msg.param.move_hero.dx=1;
      msg.param.move_hero.dy=0;
      append_control_msg(&msg);
    } else if (strcmp(info->params[0],"move_north")==0) {
      while (!msg.treated) sched_yield();
      msg.type=msg_move_hero;
      msg.response=NULL;
      msg.param.move_hero.dx=0;
      msg.param.move_hero.dy=-1;
      append_control_msg(&msg);
    } else if (strcmp(info->params[0],"move_south")==0) {
      while (!msg.treated) sched_yield();
      msg.type=msg_move_hero;
      msg.response=NULL;
      msg.param.move_hero.dx=0;
      msg.param.move_hero.dy=1;
      append_control_msg(&msg);
    } else if (strcmp(info->params[0],"write_square")==0) {
      while (!msg.treated) sched_yield();
      msg.type=msg_write_square;
      msg.response=NULL;
      msg.param.write_square.x=disp_x+info->x;
      msg.param.write_square.y=disp_y+info->y;
      if (*info->num_params>=2)
	msg.param.write_square.v=strtoul(info->params[1],NULL,10);
      else
	msg.param.write_square.v=0;
      append_control_msg(&msg);
    } else if (strcmp(info->params[0],"write_chunk")==0) {
      while (!msg.treated) sched_yield();
      msg.type=msg_write_chunk;
      msg.response=NULL;
      msg.param.write_chunk.x=disp_x+info->x;
      msg.param.write_chunk.y=disp_y+info->y;
      if (*info->num_params>=2)
	msg.param.write_square.v=strtoul(info->params[1],NULL,10);
      else
	msg.param.write_square.v=0;
      append_control_msg(&msg);
    }
  }
}

static void
init_input(void)
{
  /* Initialize thread support (Xlib and X toolkit) */
  if (XInitThreads()==0)
    lgd_fatal("This Xlib does not support threads!");
  if (XtToolkitThreadInitialize()!=True)
    lgd_fatal("This Xt does not support threads!");
  /* Use the default language procedure (?) */
  XtSetLanguageProc(NULL,(XtLanguageProc)NULL,NULL);
  /* Create the widgets */
  toplevel = XtVaAppInitialize(&app_context,APP_CLASS,
			       NULL,0,&argc,argv,
			       fallback_resources,
			       NULL);
  ground_box = XtVaCreateManagedWidget("groundBox",boxWidgetClass,
				       toplevel,
				       NULL);
  cute_label = XtVaCreateManagedWidget("cuteLabel",labelWidgetClass,
				       ground_box,
				       XtNlabel,GAMENAME,
				       NULL);
  quit_button = XtVaCreateManagedWidget("quitButton",commandWidgetClass,
					ground_box,
					NULL);
  view_area = XtVaCreateManagedWidget("viewArea",viewAreaWidgetClass,
				      ground_box,
				      XtNarrayWidthInCells,WIDTH,
				      XtNarrayHeightInCells,HEIGHT,
				      XtNarrayData,disp_tab,
				      NULL);
  /* Specify the callbacks */
  XtAddCallback(quit_button,XtNcallback,quit_callback,NULL);
  XtAddCallback(view_area,XtNcallback,view_callback,NULL);
  /* Get the ball rolling */
  XtRealizeWidget(toplevel);
  /* Bear witness of all this */
  msg.treated=1;
  msg.in_heap=0;
  input_inited=1;
}

void
redraw(void)
     /* Send an X Expose event so that the ViewArea widget will redraw
      * itself. */
{
  XEvent evt;
  evt.type=Expose;
  evt.xexpose.x=0;
  evt.xexpose.y=0;
  evt.xexpose.width=512;
  evt.xexpose.height=512;
  evt.xexpose.display=XtDisplay(view_area);
  evt.xexpose.window=XtWindow(view_area);
  evt.xexpose.count=0;
  XSendEvent(XtDisplay(view_area),XtWindow(view_area),
	     True,ExposureMask,&evt);
  XFlush(XtDisplay(view_area));
}

void *
input_main(void *garbage)
{
  init_input();
  XtAppMainLoop(app_context);
}
