/* This file is part of Legendes II */
/* Copyright (c) 1998 by
 *   Philippe Clavel (philippe.clavel@supelec.fr)
 *   David Madore (david.madore@ens.fr) */

/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
 * the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* File mapchunks.c */
/* This file contain the routines and functions, for use by the server
 * thread(s), that access the map and chunk files. */

#include <stdio.h>

#include "lgd0.h"
#include "lgdlib.h"
#include "mapchunks.h"

#define CHUNK_FILE "chunk"

lgd_filedesc_t chunkf; /* The chunk file itself */

#define CHUNK_CACHE_SIZE 16

#define CHUNK_SIZE 16 /* Size of a chunk in squares */

#define CHUNK_MASK 0xfffffff0
/* Mask to be applied to a position to obtain the position of the
 * north-westernmost corner of its chunk. */
#define CHUNK_ANTIMASK 0x0000000f
/* Mask to be applied to a position to obtain this position relative
 * to its chunk. */

typedef u_int32_t chunk_t[CHUNK_SIZE][CHUNK_SIZE];

struct chunk_cache_entry_s {
  char loaded; /* Is cache entry valid? */
  u_int32_t n; /* Chunk number */
  chunk_t c;
} chunk_cache[CHUNK_CACHE_SIZE];

static void
flush_chunk_cache(void)
{
  int i;
  for (i=0;i<CHUNK_CACHE_SIZE;i++)
    chunk_cache[i].loaded=0;
}

static int
cache_chunk(u_int32_t n)
     /* Make sure a chunk is in the chunk cache, loading it from disk
      * if necessary.  Return its entry number in the cache. */
{
  int i;
  for (i=0;i<CHUNK_CACHE_SIZE;i++) {
    if (chunk_cache[i].loaded&&(chunk_cache[i].n==n)) {
      return i;
    }
  }
  i=lgd_random(CHUNK_CACHE_SIZE);
  chunk_cache[i].loaded=1;
  chunk_cache[i].n=n;
  lgd_read_littleendian_int32(1,chunkf,(u_int32_t *)chunk_cache[i].c,
			      n*CHUNK_SIZE*CHUNK_SIZE,CHUNK_SIZE*CHUNK_SIZE);
  return i;
}

u_int32_t
read_chunk_square(u_int32_t n,long int x,long int y)
     /* Read the chunk file to see what chunk number n contains at
      * position (x,y). */
{
  return chunk_cache[cache_chunk(n)].c[x&CHUNK_ANTIMASK][y&CHUNK_ANTIMASK];
}

void
write_chunk_square(u_int32_t n,long int x,long int y,u_int32_t v)
     /* Write a square in a chunk.  Normally the chunks are read-only.
      * But some map editing programs might need this. */
{
  int i;
  i=cache_chunk(n);
  chunk_cache[i].c[x&CHUNK_ANTIMASK][y&CHUNK_ANTIMASK]=v;
  lgd_write_littleendian_int32(1,chunkf,(u_int32_t *)chunk_cache[i].c,
			       n*CHUNK_SIZE*CHUNK_SIZE,CHUNK_SIZE*CHUNK_SIZE);
}

#define MAP_FILE "map"

lgd_filedesc_t mapf; /* The map file itself */

#define ZONE_CACHE_SIZE 8

#define ZONE_SIZE_IN_CHUNKS 16 /* Size of a zone in CHUNKS */
#define ZONE_SIZE_IN_SQUARES (ZONE_SIZE_IN_CHUNKS*CHUNK_SIZE)

#define ZONE_MASK 0xffffff00
/* Mask to be applied to a position to obtain the position of the
 * north-westernmost corner of its zone. */
#define ZONE_ANTIMASK 0x000000ff
/* Mask to be applied to a position to obtain this position relative
 * to its zone. */

typedef u_int32_t zone_t[ZONE_SIZE_IN_CHUNKS][ZONE_SIZE_IN_CHUNKS];

struct zone_cache_entry_s {
  char loaded; /* Is cache entry valid? */
  int lev; /* Level of zone */
  long int x,y; /* North-westernmost corner of zone */
  /* x and y are measured in elementary squares (not in zones - not in
   * chunks). */
  zone_t z;
} zone_cache[ZONE_CACHE_SIZE];

static off_t
zone_offset(int lev,long int x,long int y)
     /* Return the position in the map file that contains the zone
      * with coordinates (x,y) (x and y refer to the zone's upper-left
      * corner). */
{
  x/=ZONE_SIZE_IN_SQUARES;
  y/=ZONE_SIZE_IN_SQUARES;
  return (lev*256+x*16+y)*ZONE_SIZE_IN_CHUNKS*ZONE_SIZE_IN_CHUNKS;
}

static void
flush_zone_cache(void)
{
  int i;
  for (i=0;i<ZONE_CACHE_SIZE;i++)
    zone_cache[i].loaded=0;
}

static int
cache_zone(int lev,long x,long y)
     /* Make sure a zone is in the zone cache, loading it from disk if
      * necessary.  Return its entry number in the cache. */
{
  int i;
  long int zx,zy;
  zx=x&ZONE_MASK;
  zy=y&ZONE_MASK;
  for (i=0;i<ZONE_CACHE_SIZE;i++) {
    if (zone_cache[i].loaded&&(zone_cache[i].lev==lev)&&
	(zone_cache[i].x==zx)&&(zone_cache[i].y==zy)) {
      return i;
    }
  }
  i=lgd_random(ZONE_CACHE_SIZE);
  zone_cache[i].loaded=1;
  zone_cache[i].lev=lev;
  zone_cache[i].x=zx;
  zone_cache[i].y=zy;
  lgd_read_littleendian_int32(1,mapf,(u_int32_t *)zone_cache[i].z,
			      zone_offset(lev,zx,zy),
			      ZONE_SIZE_IN_CHUNKS*ZONE_SIZE_IN_CHUNKS);
  return i;
}

u_int32_t
read_map_chunk(int lev,long int x,long int y)
     /* Read the map file to return the chunk that contains the square
      * at position (x,y) of level lev. */
{
  return zone_cache[cache_zone(lev,x,y)].\
    z[(x&ZONE_ANTIMASK)/CHUNK_SIZE][(y&ZONE_ANTIMASK)/CHUNK_SIZE];
}

void
write_map_chunk(int lev,long int x,long int y,u_int32_t v)
     /* Write a chunk in the map.  Normally the map is read-only.  But
      * some map editing programs might need this. */
{
  int i;
  long int zx,zy;
  zx=x&ZONE_MASK;
  zy=y&ZONE_MASK;
  i=cache_zone(lev,x,y);
  zone_cache[i].\
    z[(x&ZONE_ANTIMASK)/CHUNK_SIZE][(y&ZONE_ANTIMASK)/CHUNK_SIZE]=v;
  lgd_write_littleendian_int32(1,mapf,(u_int32_t *)zone_cache[i].z,
			       zone_offset(lev,zx,zy),
			       ZONE_SIZE_IN_CHUNKS*ZONE_SIZE_IN_CHUNKS);
}

/* These will disappear eventually */
#define MAP_SIZE_X 4096
#define MAP_SIZE_Y 4096

u_int32_t
read_map_square(int lev,long int x,long int y)
     /* This is the terminal map-reading function.  It returns the
      * value of the elementary square at position (x,y) in the
      * map. */
{
  if ((x<0)||(y<0)||(x>=MAP_SIZE_X)||(y>=MAP_SIZE_Y)) return 0;
  /* The previous line is only temporary and will disappear as soon as
   * we really have a world in place. */
  return read_chunk_square(read_map_chunk(lev,x,y),x,y);
}

void
init_map_engine(void)
{
  lgd_open(1,&mapf,MAP_FILE,OPEN_READWRITE);
  lgd_open(1,&chunkf,CHUNK_FILE,OPEN_READWRITE);
  flush_zone_cache();
  flush_chunk_cache();
}
